"""
Functions for interoperability between Q-CTRL software stack
and QuEra packages, and  utilities to assist with submitting jobs
to the QuEra Aquila QPU via Braket, and analyzing resulting data.
"""

import boulderopal as bo
import jsonpickle
import matplotlib.pyplot as plt
import numpy as np
import scipy
from juliacall import Main as jl
from juliacall import convert as jlconvert

jl.seval("using Bloqade")

from matplotlib.colors import LinearSegmentedColormap, to_rgb

density_colors = [
    "#680CE9",
    "#B482FA",
    "#EDE0FE",
    "#FFFFFF",
    "#FED6D7",
    "#FA7370",
    "#C02C21",
]
density_stops = [0.0, 1 / 3, 5 / 12, 0.5, 7 / 12, 2 / 3, 1.0]
density_cmap = LinearSegmentedColormap.from_list(
    "density_cmap", [(s, to_rgb(c)) for (s, c) in zip(density_stops, density_colors)]
)


def save_result(filename, result):
    with open(filename, "w") as f:
        f.write(jsonpickle.encode(result))
    return filename


def load_result(filename):
    with open(filename, "r") as f:
        return jsonpickle.decode(f.read())


def load_braket_result(filename):
    """
    Load a result in the format output by the save_braket_result
    or save_braket_post_processed_result functions.

    Parameters
    ----------
    filename : filename or file handle
        Path of file to load from.

    Returns
    ----------
    qpu_result : np.ndarray
        numpy array of shape (number of shots, number of qubits), giving
        the measurement results from each shot of the experiment. values
        of 1 correspond to the ground state and 0 correspond to the
        Rydberg state.
        (if the result was not post-processed, 0 values also include un-filled
        lattice sites)
    """
    return np.loadtxt(filename)


def save_braket_result(filename, result, drop_bad_shots=False):
    """
    Save shot results from a Braket job.

    params:
    filename - Name of file to save to.
    result - Amazon Braket result object containing shots and site-filling metadata.
    drop_bad_shots - If True, drop shots with unfilled lattice sites. Default False.
    """
    qpu_results = []
    for measurement in result.measurements:
        if drop_bad_shots and not measurement.pre_sequence.all():
            # reject any shot with unfilled sites
            continue
        qpu_results.append(measurement.post_sequence)

    qpu_results = np.asarray(qpu_results)
    np.savetxt(filename, qpu_results)


def _julia_convert_vector(array):
    """
    Convert lists or numpy arrays into Julia vectors of type
    Float64.

    Parameters
    ----------
    array : list or np.ndarray
        Python list or numpy array of floating-point numbers.

    Returns
    ----------
    jl.Vector
        Julia Vector object containing the same data as the input array.
    """
    return jlconvert(jl.Vector[jl.Float64], array)


def bloqade_simulate_nq(num_atoms, hamiltonian, duration, step=1e-3):
    """
    Given a Hamiltonian (in the ground-Rydberg space, not including hyperfine states)
    and a duration to simulate for, simulate this Hamiltonian on a system of neutral atoms,
    starting with all atoms in the ground state.

    Parameters
    ----------
    num_atoms : int
        number of atoms in the system, for constructing the initial state
    hamiltonian
        Hamiltonian to simulate, defined for a lattice with num_atoms atoms.
    duration : float
        duration to run the simulation for.
    step : float
        timestep for the simulation, default 1e-3. Units µs.

    Returns
    ----------
    state : np.ndarray
        final state of the atom register after the simulation.
    """
    # Set up initial (ground) state
    st = jl.zeros(jl.ComplexF64, 2**num_atoms)
    st[0] = 1.0
    reg = jl.arrayreg(st, nlevel=2)
    prob = jl.KrylovEvolution(reg, jl.range(0.0, duration, step=step), hamiltonian)
    jl.emulate_b(prob)
    return np.array(jl.state(reg))


def chain_lattice(num_atoms, atom_separation):
    """
    Generate a Bloqade AtomList for a linear chain of atoms.

    Parameters
    ----------
    num_atoms : int
        length of the chain
    atom_separation : float
        separation between adjacent atoms, units µm.

    Returns
    ----------
    lattice : Bloqade AtomList
        AtomList for linear chain matching the input parameters.
    """
    return jl.generate_sites(jl.ChainLattice(), num_atoms, scale=atom_separation)


def bloqade_pwc(values, durations):
    """
    Create a Bloqade PWC signal from lists of values and durations.

    Parameters
    ----------
    values : np.ndarray
        PWC values.
    durations : np.ndarray
        durations of PWC segments. Units µs (Bloqade native units are MHz/µs).

    returns:
    pwc : Bloqade Waveform
        the Bloqade-compatible PWC signal.
    """
    clocks = _julia_convert_vector(np.concatenate(([0], np.cumsum(durations))))
    values = _julia_convert_vector(values)

    return jl.piecewise_constant(clocks=clocks, values=values)


def bloqade_two_level_hamiltonian(atoms, omega_r=None, delta_r=None, phi_r=None):
    """
    Generate a Bloqade two-level Hamiltonian from signals for Rabi rate, detuning
    and phase, for a pulse on the Rydberg transition.

    Parameters
    ----------
    omega_r : Bloqade Waveform or float, optional
        Rabi rate signal for the pulse, as a Bloqade waveform or float (if constant).
    delta_r : Bloqade Waveform or float, optional
        Detuning signal for the pulse, as a Bloqade waveform or float (if constant).
    phi_r : Bloqade Waveform or float, optional
        Phase modulation signal for the pulse, as a Bloqade waveform or float (if constant).

    Returns
    ----------
    hamiltonian
        The Bloqade Hamiltonian constructed with the input parameters.
    """
    params_dict = {}

    if omega_r is not None:
        params_dict["Ω"] = omega_r
    if delta_r is not None:
        params_dict["Δ"] = delta_r
    if phi_r is not None:
        params_dict["ϕ"] = phi_r

    return jl.rydberg_h(atoms, **params_dict)


def qctrl_pulse_to_bloqade(result):
    """
    Take a Boulder Opal result dictionary from optimizing a pulse, and convert it to Bloqade PWCs
    for the amplitude and detuning signals in the pulse.

    Parameters
    ----------
    result : dict
        Result dictionary from Boulder Opal.

    Returns:
    omega : Bloqade Waveform
        Amplitude PWC, as Bloqade Waveform.
    delta : Bloqade Waveform
        Detuning PWC, as Bloqade Waveform.
    """
    # Extract PWC segments and convert to Bloqade's native MHz/µs units
    amplitude_pwc_values = np.copy(result["output"]["amplitude"]["values"])
    amplitude_pwc_durations = np.copy(result["output"]["amplitude"]["durations"])
    amplitude_pwc_durations *= 1e6
    amplitude_pwc_values /= 1e6

    detuning_pwc_values = np.copy(result["output"]["detuning"]["values"])
    detuning_pwc_durations = np.copy(result["output"]["detuning"]["durations"])
    detuning_pwc_durations *= 1e6
    detuning_pwc_values /= 1e6

    omega = bloqade_pwc(amplitude_pwc_values, amplitude_pwc_durations)
    delta = bloqade_pwc(detuning_pwc_values, detuning_pwc_durations)

    return omega, delta


def get_operators(atom_count, interaction_strength):
    h_omega_values = []
    h_omega_indices = []
    h_delta_values = []
    h_delta_indices = []
    h_fixed_values = []
    h_fixed_indices = []

    # For each computational basis state
    for j in range(2**atom_count):
        # Get binary representation
        s = np.binary_repr(j, width=atom_count)
        state = np.array([int(m) for m in s])

        # Global shift term (big delta)
        n_exc = np.sum(state)  # Number of excited states
        h_delta_values.append(n_exc)
        h_delta_indices.append([j, j])

        # omega term
        for i in range(atom_count):
            # Couple "state" and "state with qubit i flipped"
            state_p = state.copy()
            state_p[i] = 1 - state_p[i]
            h_omega_values.append(0.5)
            h_omega_indices.append([j, state_p.dot(2 ** np.arange(atom_count)[::-1])])

        h_fixed = 0.0
        # Interaction term
        for atom_1 in range(atom_count):
            for atom_2 in range(atom_1 + 1, atom_count):
                # implement interactions across the end of the chain
                # distance is the minimum of "forwards" and
                # "backwards" distance
                if state[atom_1] and state[atom_2]:
                    d_12 = atom_2 - atom_1
                    h_fixed += interaction_strength / d_12**6

        if h_fixed != 0.0:
            h_fixed_values.append(h_fixed)
            h_fixed_indices.append([j, j])

    h_omega = scipy.sparse.coo_matrix(
        (h_omega_values, np.array(h_omega_indices).T),
        shape=[2**atom_count, 2**atom_count],
    )

    h_delta = scipy.sparse.coo_matrix(
        (h_delta_values, np.array(h_delta_indices).T),
        shape=[2**atom_count, 2**atom_count],
    )

    h_fixed = scipy.sparse.coo_matrix(
        (h_fixed_values, np.array(h_fixed_indices).T),
        shape=[2**atom_count, 2**atom_count],
    )

    return h_omega.todense(), h_delta.todense(), h_fixed.todense()


def _get_target_state(atom_count):
    idx_even = sum(2**n for n in range(0, atom_count, 2))
    idx_odd = sum(2**n for n in range(1, atom_count, 2))

    target_state = np.zeros([2**atom_count])
    target_state[[idx_even, idx_odd]] = 1.0 / np.sqrt(2.0)
    return target_state


def average_rydberg_density(state, atom_count):
    """
    Get a the average Rydberg density for each atom given a state vector.

    Parameters
    ----------
    state : np.ndarray
        The state vector to calculate Rydberg density for.
    atom_count : int
        Number of atoms in the system the state vector represents.

    Returns
    ----------
    rydberg_density : np.ndarray
        1D array containing the average Rydberg density for each atom.
    """
    rydberg_density = np.zeros(atom_count)
    for i in range(atom_count):
        include = (np.arange(2**atom_count) // (2**i)) % 2 == 1
        rydberg_density[i] = np.sum(np.abs(state[include]) ** 2)
    return rydberg_density


def pwc_to_piecewise_linear(Omega_pwc, Delta_pwc, min_interval=50e-9):
    """
    (approximately) convert a piecewise-constant pulse (e.g. Boulder optimization output)
    to a piecewise-linear pulse, required for drive specification
    on QuEra device.

    Parameters
    ----------
    Omega_pwc : dict of np.ndarray
        input PWC pulse specification for Rabi rate, as a dictionary containing
        a key for "values" and "durations". Units rad/s
    Delta_pwc : dict of np.ndarray
        input PWC pulse specification for detuning, in same format and units as Omega_pwc
    min_interval : float, optional
        minimum sampling interval to use for the piecewise-linear output. Units seconds,
        default 50e-9

    Returns
    ----------
    pwl_clocks : np.ndarray
        time sampling points for the piecewise-linear pulse. Units seconds
    omega_values : np.ndarray
        value of Rabi rate at each sampling point. Units rad/s
    delta_values : np.ndarray
        value of detuning at each sampling point. Units rad/s
    """

    seg_durations = Omega_pwc["durations"]
    clocks_end = np.cumsum(seg_durations)

    total_duration = np.sum(seg_durations)
    num_points = int(total_duration // min_interval) + 1

    pwl_clocks = np.linspace(0, total_duration, num_points)

    omega_values = np.zeros_like(pwl_clocks)
    delta_values = np.zeros_like(pwl_clocks)

    seg_idx = 0
    for i, t in enumerate(pwl_clocks):
        # possible we need to skip some segments due to coarse time resolution of quera device
        while t > clocks_end[seg_idx] and seg_idx < len(Omega_pwc["values"]) - 1:
            seg_idx += 1
        omega_values[i] = Omega_pwc["values"][seg_idx]
        delta_values[i] = Delta_pwc["values"][seg_idx]

        if i == 0 or i == num_points - 1:
            omega_values[i] = 0.0

    # do some rounding maths to force the time points and control values
    # to conform to spec for QuEra
    pwl_clocks = np.round(pwl_clocks, 9)

    omega_values = np.round(omega_values / 400, 0) * 400

    delta_values = np.round(np.round(delta_values / 0.2, 0) * 0.2, 1)

    return pwl_clocks, omega_values, delta_values


def plot_bitstring_comparison(pmf_1, pmf_2, axes, label_1="", label_2="", num_bars=10):
    num_atoms = int(np.log2(len(pmf_1)))

    target_bitstring = "10" * (num_atoms // 2) + "1"

    xs = np.argsort(pmf_1)[::-1][:num_bars]
    is_target = np.array([np.binary_repr(x, num_atoms) for x in xs]) == target_bitstring

    probs_1 = pmf_1[xs]
    probs_2 = pmf_2[xs]
    probs_ideal = 1.0 * is_target

    plt.sca(axes)

    plt.bar(np.arange(num_bars), probs_1, width=0.3, label=label_1)
    plt.bar(np.arange(num_bars) + 0.3, probs_2, width=0.3, label=label_2)
    plt.bar(
        np.arange(num_bars) - 0.3, probs_ideal, width=0.3, label="Ideal", color="grey"
    )

    plt.xticks(
        np.arange(num_bars),
        labels=[np.binary_repr(x, num_atoms) for x in xs],
        rotation=45,
        ha="right",
    )
    plt.ylim([0, 1])
    plt.xlabel("Bitstring")
    plt.ylabel("Probability")

    for i in range(num_bars):
        if is_target[i]:
            axes.get_xticklabels()[i].set_color("red")


def add_pwc_signals_to_graph(
    graph, optimization_variable_count, omega_max, delta_max, duration
):
    amplitude = graph.real_optimizable_pwc_signal(
        segment_count=optimization_variable_count, maximum=omega_max, duration=duration
    )
    detuning = graph.real_optimizable_pwc_signal(
        segment_count=optimization_variable_count,
        maximum=delta_max,
        minimum=-delta_max,
        duration=duration,
    )

    return amplitude, detuning


def envelope_and_filter_graph(
    graph,
    segment_count,
    amplitude,
    detuning,
    filter_kernel_width,
    flat_fraction,
    duration,
):
    flat_gaussian_envelope = graph.signals.gaussian_pulse_pwc(
        duration=duration,
        segment_count=segment_count,
        amplitude=1,
        flat_duration=duration * flat_fraction,
    )

    kernel = graph.gaussian_convolution_kernel(filter_kernel_width)

    omega = graph.filter_and_resample_pwc(
        amplitude, kernel=kernel, segment_count=segment_count, duration=duration
    )
    omega = omega * flat_gaussian_envelope

    delta = graph.filter_and_resample_pwc(
        detuning, kernel=kernel, segment_count=segment_count
    )

    return omega, delta
